"""Custom accessor functionality for modin.

Source code adapted from pyspark.pandas implementation:
https://spark.apache.org/docs/3.2.0/api/python/reference/pyspark.pandas/api/pyspark.pandas.extensions.register_dataframe_accessor.html?highlight=register_dataframe_accessor#pyspark.pandas.extensions.register_dataframe_accessor
"""

import warnings

from pandera.accessors.pandas_accessor import (
    PanderaDataFrameAccessor,
    PanderaSeriesAccessor,
)


# pylint: disable=too-few-public-methods
class CachedAccessor:
    """
    Custom property-like object.

    A descriptor for caching accessors:

    :param name: Namespace that accessor's methods, properties, etc will be
        accessed under, e.g. "foo" for a dataframe accessor yields the accessor
        ``df.foo``
    :param cls: Class with the extension methods.

    For accessor, the class's __init__ method assumes that you are registering
    an accessor for one of ``Series``, ``DataFrame``, or ``Index``.
    """

    def __init__(self, name, accessor):
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:  # pragma: no cover
            return self._accessor
        accessor_obj = self._accessor(obj)
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _register_accessor(name, cls):
    """
    Register a custom accessor on {class} objects.

    :param name: Name under which the accessor should be registered. A warning
        is issued if this name conflicts with a preexisting attribute.
    :returns: A class decorator callable.
    """

    def decorator(accessor):
        if hasattr(cls, name):
            msg = (
                f"registration of accessor {accessor} under name '{name}' for "
                "type {cls.__name__} is overriding a preexisting attribute "
                "with the same name."
            )

            warnings.warn(
                msg,
                UserWarning,
                stacklevel=2,
            )
        setattr(cls, name, CachedAccessor(name, accessor))
        return accessor

    return decorator


def register_dataframe_accessor(name):
    """
    Register a custom accessor with a DataFrame

    :param name: name used when calling the accessor after its registered
    :returns: a class decorator callable.
    """
    # pylint: disable=import-outside-toplevel
    from modin.pandas import DataFrame

    return _register_accessor(name, DataFrame)


def register_series_accessor(name):
    """
    Register a custom accessor with a Series object

    :param name: name used when calling the accessor after its registered
    :returns: a callable class decorator
    """
    # pylint: disable=import-outside-toplevel
    from modin.pandas import Series

    return _register_accessor(name, Series)


register_dataframe_accessor("pandera")(PanderaDataFrameAccessor)
register_series_accessor("pandera")(PanderaSeriesAccessor)



"""Base classes for parsing, validation, and error Reporting Backends.

These classes implement a common interface of operations needed for
data validation. These operations are exposed as methods that are composed
together to implement the pandera schema specification.
"""

from abc import ABC
from typing import Any, Dict, List, NamedTuple, Optional, Union

# from pandera.api.base.checks import BaseCheck
from pandera.errors import SchemaError, SchemaErrorReason


class CoreCheckResult(NamedTuple):
    """Namedtuple for holding results of core checks."""

    passed: bool
    check: Union[str, "BaseCheck"]  # type: ignore
    check_index: Optional[int] = None
    check_output: Optional[Any] = None
    reason_code: Optional[SchemaErrorReason] = None
    message: Optional[str] = None
    failure_cases: Optional[Any] = None
    schema_error: Optional[SchemaError] = None
    original_exc: Optional[Exception] = None


class CoreParserResult(NamedTuple):
    """Namedtuple for holding core parser results."""


class BaseSchemaBackend(ABC):
    """Abstract base class for a schema backend implementation."""

    def preprocess(
        self,
        check_obj,
        inplace: bool = False,
    ):
        """Preprocesses a check object before applying check functions."""
        raise NotImplementedError

    def subsample(
        self,
        check_obj,
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
    ):
        """Subsamples a check object before applying check functions."""
        raise NotImplementedError

    def validate(
        self,
        check_obj,
        schema,
        *,
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
        lazy: bool = False,
        inplace: bool = False,
    ):
        """
        Parse and validate a check object, returning type-coerced and validated
        object.
        """
        raise NotImplementedError

    def coerce_dtype(
        self,
        check_obj,
        schema=None,
    ):
        """Coerce the data type of the check object."""
        raise NotImplementedError

    def run_check(
        self,
        check_obj,
        schema,
        check,
        check_index: int,
        *args,
    ):
        """Run a single check on the check object."""
        raise NotImplementedError

    def run_checks(self, check_obj, schema):
        """Run a list of checks on the check object."""
        raise NotImplementedError

    def run_schema_component_checks(
        self,
        check_obj,
        schema_components,
        lazy,
    ):
        """Run checks for all schema components."""
        raise NotImplementedError

    def check_name(self, check_obj, schema):
        """Core check that checks the name of the check object."""
        raise NotImplementedError

    def check_nullable(self, check_obj, schema):
        """Core check that checks the nullability of a check object."""
        raise NotImplementedError

    def check_unique(self, check_obj, schema):
        """Core check that checks the uniqueness of values in a check object."""
        raise NotImplementedError

    def check_dtype(self, check_obj, schema):
        """Core check that checks the data type of a check object."""
        raise NotImplementedError

    def failure_cases_metadata(
        self, schema_name: str, schema_errors: List[SchemaError]
    ):
        """Get failure cases metadata for lazy validation."""
        raise NotImplementedError


class BaseCheckBackend(ABC):
    """Abstract base class for a check backend implementation."""

    def __init__(self, check):  # pylint: disable=unused-argument
        """Initializes a check backend object."""

    def __call__(self, check_obj, key=None):
        raise NotImplementedError

    def query(self, check_obj):
        """Implements querying behavior to produce subset of check object."""
        raise NotImplementedError

    def groupby(self, check_obj):
        """Implements groupby behavior for check object."""
        raise NotImplementedError

    def aggregate(self, check_obj):
        """Implements aggregation behavior for check object."""
        raise NotImplementedError

    def preprocess(self, check_obj, key):
        """Preprocesses a check object before applying the check function."""
        raise NotImplementedError

    def postprocess(self, check_obj, check_output):
        """Postprocesses the result of applying the check function."""
        raise NotImplementedError

    def apply(self, check_obj):
        """Apply the check function to a check object."""
        raise NotImplementedError

    def statistics(self):
        """Check statistics property."""
        raise NotImplementedError

    def strategy(self):
        """Return a data generation strategy."""
        raise NotImplementedError



"""Pandas implementation of built-in checks"""

import operator
import re
from typing import cast, Any, Iterable, TypeVar, Union

import pandas as pd

import pandera.strategies as st
from pandera.api.extensions import register_builtin_check

from pandera.typing.modin import MODIN_INSTALLED
from pandera.typing.pyspark import PYSPARK_INSTALLED


if MODIN_INSTALLED and not PYSPARK_INSTALLED:  # pragma: no cover
    import modin.pandas as mpd

    PandasData = Union[pd.Series, pd.DataFrame, mpd.Series, mpd.DataFrame]
elif not MODIN_INSTALLED and PYSPARK_INSTALLED:  # pragma: no cover
    import pyspark.pandas as ppd

    PandasData = Union[pd.Series, pd.DataFrame, ppd.Series, ppd.DataFrame]  # type: ignore[misc]
elif MODIN_INSTALLED and PYSPARK_INSTALLED:  # pragma: no cover
    import modin.pandas as mpd
    import pyspark.pandas as ppd

    PandasData = Union[  # type: ignore[misc]
        pd.Series,
        pd.DataFrame,
        mpd.Series,
        mpd.DataFrame,
        ppd.Series,
        ppd.DataFrame,
    ]
else:  # pragma: no cover
    PandasData = Union[pd.Series, pd.DataFrame]  # type: ignore[misc]


T = TypeVar("T")


@register_builtin_check(
    aliases=["eq"],
    strategy=st.eq_strategy,
    error="equal_to({value})",
)
def equal_to(data: PandasData, value: Any) -> PandasData:
    """Ensure all elements of a data container equal a certain value.

    :param value: values in this pandas data structure must be
        equal to this value.
    """
    return data == value


@register_builtin_check(
    aliases=["ne"],
    strategy=st.ne_strategy,
    error="not_equal_to({value})",
)
def not_equal_to(data: PandasData, value: Any) -> PandasData:
    """Ensure no elements of a data container equals a certain value.

    :param value: This value must not occur in the checked
        :class:`pandas.Series`.
    """
    return data != value


@register_builtin_check(
    aliases=["gt"],
    strategy=st.gt_strategy,
    error="greater_than({min_value})",
)
def greater_than(data: PandasData, min_value: Any) -> PandasData:
    """
    Ensure values of a data container are strictly greater than a minimum
    value.

    :param min_value: Lower bound to be exceeded. Must be a type comparable
        to the dtype of the :class:`pandas.Series` to be validated (e.g. a
        numerical type for float or int and a datetime for datetime).
    """
    return data > min_value


@register_builtin_check(
    aliases=["ge"],
    strategy=st.ge_strategy,
    error="greater_than_or_equal_to({min_value})",
)
def greater_than_or_equal_to(data: PandasData, min_value: Any) -> PandasData:
    """Ensure all values are greater or equal a certain value.

    :param min_value: Allowed minimum value for values of a series. Must be
        a type comparable to the dtype of the :class:`pandas.Series` to be
        validated.
    """
    return data >= min_value


@register_builtin_check(
    aliases=["lt"],
    strategy=st.lt_strategy,
    error="less_than({max_value})",
)
def less_than(data: PandasData, max_value: Any) -> PandasData:
    """Ensure values of a series are strictly below a maximum value.

    :param max_value: All elements of a series must be strictly smaller
        than this. Must be a type comparable to the dtype of the
        :class:`pandas.Series` to be validated.
    """
    if max_value is None:
        raise ValueError("max_value must not be None")
    return data < max_value


@register_builtin_check(
    aliases=["le"],
    strategy=st.le_strategy,
    error="less_than_or_equal_to({max_value})",
)
def less_than_or_equal_to(data: PandasData, max_value: Any) -> PandasData:
    """Ensure values of a series are strictly below a maximum value.

    :param max_value: Upper bound not to be exceeded. Must be a type
        comparable to the dtype of the :class:`pandas.Series` to be
        validated.
    """
    if max_value is None:
        raise ValueError("max_value must not be None")
    return data <= max_value


@register_builtin_check(
    aliases=["between"],
    strategy=st.in_range_strategy,
    error="in_range({min_value}, {max_value})",
)
def in_range(
    data: PandasData,
    min_value: T,
    max_value: T,
    include_min: bool = True,
    include_max: bool = True,
):
    """Ensure all values of a series are within an interval.

    Both endpoints must be a type comparable to the dtype of the
    :class:`pandas.Series` to be validated.

    :param min_value: Left / lower endpoint of the interval.
    :param max_value: Right / upper endpoint of the interval. Must not be
        smaller than min_value.
    :param include_min: Defines whether min_value is also an allowed value
        (the default) or whether all values must be strictly greater than
        min_value.
    :param include_max: Defines whether min_value is also an allowed value
        (the default) or whether all values must be strictly smaller than
        max_value.
    """
    # Using functions from operator module to keep conditions out of the
    # closure
    left_op = operator.le if include_min else operator.lt
    right_op = operator.ge if include_max else operator.gt
    return left_op(min_value, data) & right_op(max_value, data)  # type: ignore


@register_builtin_check(
    strategy=st.isin_strategy,
    error="isin({allowed_values})",
)
def isin(data: PandasData, allowed_values: Iterable) -> PandasData:
    """Ensure only allowed values occur within a series.

    This checks whether all elements of a :class:`pandas.Series`
    are part of the set of elements of allowed values. If allowed
    values is a string, the set of elements consists of all distinct
    characters of the string. Thus only single characters which occur
    in allowed_values at least once can meet this condition. If you
    want to check for substrings use :meth:`Check.str_contains`.

    :param allowed_values: The set of allowed values. May be any iterable.
    :param kwargs: key-word arguments passed into the `Check` initializer.
    """
    return data.isin(allowed_values)


@register_builtin_check(
    strategy=st.notin_strategy,
    error="notin({forbidden_values})",
)
def notin(data: PandasData, forbidden_values: Iterable) -> PandasData:
    """Ensure some defined values don't occur within a series.

    Like :meth:`Check.isin` this check operates on single characters if
    it is applied on strings. If forbidden_values is a string, it is understood
    as set of prohibited characters. Any string of length > 1 can't be in it by
    design.

    :param forbidden_values: The set of values which should not occur. May
        be any iterable.
    :param raise_warning: if True, check raises UserWarning instead of
        SchemaError on validation.
    """
    return ~data.isin(forbidden_values)


@register_builtin_check(
    strategy=st.str_matches_strategy,
    error="str_matches('{pattern}')",
)
def str_matches(
    data: PandasData,
    pattern: Union[str, re.Pattern],
) -> PandasData:
    """Ensure that string values match a regular expression.

    :param pattern: Regular expression pattern to use for matching
    :param kwargs: key-word arguments passed into the `Check` initializer.
    """
    return data.str.match(cast(str, pattern), na=False)


@register_builtin_check(
    strategy=st.str_contains_strategy,
    error="str_contains('{pattern}')",
)
def str_contains(
    data: PandasData,
    pattern: Union[str, re.Pattern],
) -> PandasData:
    """Ensure that a pattern can be found within each row.

    :param pattern: Regular expression pattern to use for searching
    :param kwargs: key-word arguments passed into the `Check` initializer.
    """
    return data.str.contains(cast(str, pattern), na=False)


@register_builtin_check(
    strategy=st.str_startswith_strategy,
    error="str_startswith('{string}')",
)
def str_startswith(data: PandasData, string: str) -> PandasData:
    """Ensure that all values start with a certain string.

    :param string: String all values should start with
    :param kwargs: key-word arguments passed into the `Check` initializer.
    """
    return data.str.startswith(string, na=False)


@register_builtin_check(
    strategy=st.str_endswith_strategy, error="str_endswith('{string}')"
)
def str_endswith(data: PandasData, string: str) -> PandasData:
    """Ensure that all values end with a certain string.

    :param string: String all values should end with
    :param kwargs: key-word arguments passed into the `Check` initializer.
    """
    return data.str.endswith(string, na=False)


@register_builtin_check(
    strategy=st.str_length_strategy,
    error="str_length({min_value}, {max_value})",
)
def str_length(
    data: PandasData,
    min_value: int = None,
    max_value: int = None,
) -> PandasData:
    """Ensure that the length of strings is within a specified range.

    :param min_value: Minimum length of strings (default: no minimum)
    :param max_value: Maximum length of strings (default: no maximum)
    """
    str_len = data.str.len()
    if min_value is None and max_value is None:
        raise ValueError(
            "At least a minimum or a maximum need to be specified. Got "
            "None."
        )
    if max_value is None:
        return str_len >= min_value  # type: ignore[operator]
    elif min_value is None:
        return str_len <= max_value
    return (str_len <= max_value) & (str_len >= min_value)


@register_builtin_check(
    error="unique_values_eq({values})",
)
def unique_values_eq(data: PandasData, values: Iterable):
    """Ensure that unique values in the data object contain all values.

    .. note::
        In constrast with :func:`isin`, this check makes sure that all the items
        in the ``values`` iterable are contained within the series.

    :param values: The set of values that must be present. Maybe any iterable.
    """
    return set(data.unique()) == values  # type: ignore[return-value,operator]



"""Core pandas schema component specifications."""

import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast

import pandas as pd

import pandera.strategies as st
from pandera import errors
from pandera.api.pandas.array import ArraySchema
from pandera.api.pandas.container import DataFrameSchema
from pandera.api.pandas.types import CheckList, PandasDtypeInputTypes
from pandera.dtypes import UniqueSettings


class Column(ArraySchema):
    """Validate types and properties of DataFrame columns."""

    def __init__(
        self,
        dtype: PandasDtypeInputTypes = None,
        checks: Optional[CheckList] = None,
        nullable: bool = False,
        unique: bool = False,
        report_duplicates: UniqueSettings = "all",
        coerce: bool = False,
        required: bool = True,
        name: Union[str, Tuple[str, ...], None] = None,
        regex: bool = False,
        title: Optional[str] = None,
        description: Optional[str] = None,
        default: Optional[Any] = None,
    ) -> None:
        """Create column validator object.

        :param dtype: datatype of the column. The datatype for type-checking
            a dataframe. If a string is specified, then assumes
            one of the valid pandas string values:
            http://pandas.pydata.org/pandas-docs/stable/basics.html#dtypes
        :param checks: checks to verify validity of the column
        :param nullable: Whether or not column can contain null values.
        :param unique: whether column values should be unique
        :param report_duplicates: how to report unique errors
            - `exclude_first`: report all duplicates except first occurence
            - `exclude_last`: report all duplicates except last occurence
            - `all`: (default) report all duplicates
        :param coerce: If True, when schema.validate is called the column will
            be coerced into the specified dtype. This has no effect on columns
            where ``dtype=None``.
        :param required: Whether or not column is allowed to be missing
        :param name: column name in dataframe to validate.
        :param regex: whether the ``name`` attribute should be treated as a
            regex pattern to apply to multiple columns in a dataframe.
        :param title: A human-readable label for the column.
        :param description: An arbitrary textual description of the column.
        :param default: The default value for missing values in the column.

        :raises SchemaInitError: if impossible to build schema from parameters

        :example:

        >>> import pandas as pd
        >>> import pandera as pa
        >>>
        >>>
        >>> schema = pa.DataFrameSchema({
        ...     "column": pa.Column(str)
        ... })
        >>>
        >>> schema.validate(pd.DataFrame({"column": ["foo", "bar"]}))
          column
        0    foo
        1    bar

        See :ref:`here<column>` for more usage details.
        """
        super().__init__(
            dtype=dtype,
            checks=checks,
            nullable=nullable,
            unique=unique,
            report_duplicates=report_duplicates,
            coerce=coerce,
            name=name,
            title=title,
            description=description,
            default=default,
        )
        if (
            name is not None
            and not isinstance(name, str)
            and not is_valid_multiindex_key(name)
            and regex
        ):
            raise ValueError(
                "You cannot specify a non-string name when setting regex=True"
            )
        self.required = required
        self.name = name
        self.regex = regex

    @property
    def _allow_groupby(self) -> bool:
        """Whether the schema or schema component allows groupby operations."""
        return True

    @property
    def properties(self) -> Dict[str, Any]:
        """Get column properties."""
        return {
            "dtype": self.dtype,
            "checks": self.checks,
            "nullable": self.nullable,
            "unique": self.unique,
            "report_duplicates": self.report_duplicates,
            "coerce": self.coerce,
            "required": self.required,
            "name": self.name,
            "regex": self.regex,
            "title": self.title,
            "description": self.description,
            "default": self.default,
        }

    def set_name(self, name: str):
        """Used to set or modify the name of a column object.

        :param str name: the name of the column object

        """
        self.name = name
        return self

    def validate(
        self,
        check_obj: pd.DataFrame,
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
        lazy: bool = False,
        inplace: bool = False,
    ) -> pd.DataFrame:
        """Validate a Column in a DataFrame object.

        :param check_obj: pandas DataFrame to validate.
        :param head: validate the first n rows. Rows overlapping with `tail` or
            `sample` are de-duplicated.
        :param tail: validate the last n rows. Rows overlapping with `head` or
            `sample` are de-duplicated.
        :param sample: validate a random sample of n rows. Rows overlapping
            with `head` or `tail` are de-duplicated.
        :param random_state: random seed for the ``sample`` argument.
        :param lazy: if True, lazily evaluates dataframe against all validation
            checks and raises a ``SchemaErrors``. Otherwise, raise
            ``SchemaError`` as soon as one occurs.
        :param inplace: if True, applies coercion to the object of validation,
            otherwise creates a copy of the data.
        :returns: validated DataFrame.
        """
        return self.get_backend(check_obj).validate(
            check_obj,
            self,
            head=head,
            tail=tail,
            sample=sample,
            random_state=random_state,
            lazy=lazy,
            inplace=inplace,
        )

    def get_regex_columns(
        self, columns: Union[pd.Index, pd.MultiIndex]
    ) -> Iterable:
        """Get matching column names based on regex column name pattern.

        :param columns: columns to regex pattern match
        :returns: matchin columns
        """
        # pylint: disable=import-outside-toplevel
        from pandera.backends.pandas.components import ColumnBackend

        return cast(
            ColumnBackend, self.get_backend(pd.DataFrame())
        ).get_regex_columns(self, columns)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        def _compare_dict(obj):
            return {
                k: v if k != "_checks" else set(v)
                for k, v in obj.__dict__.items()
            }

        return _compare_dict(self) == _compare_dict(other)

    ############################
    # Schema Transform Methods #
    ############################

    @st.strategy_import_error
    def strategy(self, *, size=None):
        """Create a ``hypothesis`` strategy for generating a Column.

        :param size: number of elements to generate
        :returns: a dataframe strategy for a single column.
        """
        return super().strategy(size=size).map(lambda x: x.to_frame())

    @st.strategy_import_error
    def strategy_component(self):
        """Generate column data object for use by DataFrame strategy."""
        return st.column_strategy(
            self.dtype,
            checks=self.checks,
            unique=self.unique,
            name=self.name,
        )

    def example(self, size=None) -> pd.DataFrame:
        """Generate an example of a particular size.

        :param size: number of elements in the generated Index.
        :returns: pandas DataFrame object.
        """
        # pylint: disable=import-outside-toplevel,cyclic-import,import-error
        import hypothesis

        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore",
                category=hypothesis.errors.NonInteractiveExampleWarning,
            )
            return (
                super()
                .strategy(size=size)
                .example()
                .rename(self.name)
                .to_frame()
            )


class Index(ArraySchema):
    """Validate types and properties of a DataFrame Index."""

    @property
    def names(self):
        """Get index names in the Index schema component."""
        return [self.name]

    @property
    def _allow_groupby(self) -> bool:
        """Whether the schema or schema component allows groupby operations."""
        return False

    def validate(
        self,
        check_obj: Union[pd.DataFrame, pd.Series],
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
        lazy: bool = False,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, pd.Series]:
        """Validate DataFrameSchema or SeriesSchema Index.

        :check_obj: pandas DataFrame of Series containing index to validate.
        :param head: validate the first n rows. Rows overlapping with `tail` or
            `sample` are de-duplicated.
        :param tail: validate the last n rows. Rows overlapping with `head` or
            `sample` are de-duplicated.
        :param sample: validate a random sample of n rows. Rows overlapping
            with `head` or `tail` are de-duplicated.
        :param random_state: random seed for the ``sample`` argument.
        :param lazy: if True, lazily evaluates dataframe against all validation
            checks and raises a ``SchemaErrors``. Otherwise, raise
            ``SchemaError`` as soon as one occurs.
        :param inplace: if True, applies coercion to the object of validation,
            otherwise creates a copy of the data.
        :returns: validated DataFrame or Series.
        """
        return self.get_backend(check_obj).validate(
            check_obj,
            self,
            head=head,
            tail=tail,
            sample=sample,
            random_state=random_state,
            lazy=lazy,
            inplace=inplace,
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    ###########################
    # Schema Strategy Methods #
    ###########################

    @st.strategy_import_error
    def strategy(self, *, size: int = None):
        """Create a ``hypothesis`` strategy for generating an Index.

        :param size: number of elements to generate.
        :returns: index strategy.
        """
        return st.index_strategy(
            self.dtype,  # type: ignore
            checks=self.checks,
            nullable=self.nullable,
            unique=self.unique,
            name=self.name,
            size=size,
        )

    @st.strategy_import_error
    def strategy_component(self):
        """Generate column data object for use by MultiIndex strategy."""
        return st.column_strategy(
            self.dtype,
            checks=self.checks,
            unique=self.unique,
            name=self.name,
        )

    def example(self, size: int = None) -> pd.Index:
        """Generate an example of a particular size.

        :param size: number of elements in the generated Index.
        :returns: pandas Index object.
        """
        # pylint: disable=import-outside-toplevel,cyclic-import,import-error
        import hypothesis

        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore",
                category=hypothesis.errors.NonInteractiveExampleWarning,
            )
            return self.strategy(size=size).example()


class MultiIndex(DataFrameSchema):
    """Validate types and properties of a DataFrame MultiIndex.

    This class inherits from :class:`~pandera.api.pandas.container.DataFrameSchema` to
    leverage its validation logic.
    """

    def __init__(
        self,
        indexes: List[Index],
        coerce: bool = False,
        strict: bool = False,
        name: str = None,
        ordered: bool = True,
        unique: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """Create MultiIndex validator.

        :param indexes: list of Index validators for each level of the
            MultiIndex index.
        :param coerce: Whether or not to coerce the MultiIndex to the
            specified dtypes before validation
        :param strict: whether or not to accept columns in the MultiIndex that
            aren't defined in the ``indexes`` argument.
        :param name: name of schema component
        :param ordered: whether or not to validate the indexes order.
        :param unique: a list of index names that should be jointly unique.

        :example:

        >>> import pandas as pd
        >>> import pandera as pa
        >>>
        >>>
        >>> schema = pa.DataFrameSchema(
        ...     columns={"column": pa.Column(int)},
        ...     index=pa.MultiIndex([
        ...         pa.Index(str,
        ...               pa.Check(lambda s: s.isin(["foo", "bar"])),
        ...               name="index0"),
        ...         pa.Index(int, name="index1"),
        ...     ])
        ... )
        >>>
        >>> df = pd.DataFrame(
        ...     data={"column": [1, 2, 3]},
        ...     index=pd.MultiIndex.from_arrays(
        ...         [["foo", "bar", "foo"], [0, 1, 2]],
        ...         names=["index0", "index1"],
        ...     )
        ... )
        >>>
        >>> schema.validate(df)
                       column
        index0 index1
        foo    0            1
        bar    1            2
        foo    2            3

        See :ref:`here<multiindex>` for more usage details.

        """
        if any(not isinstance(i, Index) for i in indexes):
            raise errors.SchemaInitError(
                f"expected a list of Index objects, found {indexes} "
                f"of type {[type(x) for x in indexes]}"
            )
        self.indexes = indexes
        columns = {}
        for i, index in enumerate(indexes):
            if not ordered and index.name is None:
                # if the MultiIndex is not ordered, there's no way of
                # determining how to get the index level without an explicit
                # index name
                raise errors.SchemaInitError(
                    "You must specify index names if MultiIndex schema "
                    "component is not ordered."
                )
            columns[i if index.name is None else index.name] = Column(
                dtype=index._dtype,
                checks=index.checks,
                nullable=index.nullable,
                unique=index.unique,
            )
        super().__init__(
            columns=columns,
            coerce=coerce,
            strict=strict,
            name=name,
            ordered=ordered,
            unique=unique,
        )

    @property
    def names(self):
        """Get index names in the MultiIndex schema component."""
        return [index.name for index in self.indexes]

    @property
    def coerce(self):
        """Whether or not to coerce data types."""
        return self._coerce or any(index.coerce for index in self.indexes)

    @coerce.setter
    def coerce(self, value: bool) -> None:
        """Set coerce attribute."""
        self._coerce = value

    def validate(  # type: ignore
        self,
        check_obj: Union[pd.DataFrame, pd.Series],
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
        lazy: bool = False,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, pd.Series]:
        """Validate DataFrame or Series MultiIndex.

        :param check_obj: pandas DataFrame of Series to validate.
        :param head: validate the first n rows. Rows overlapping with `tail` or
            `sample` are de-duplicated.
        :param tail: validate the last n rows. Rows overlapping with `head` or
            `sample` are de-duplicated.
        :param sample: validate a random sample of n rows. Rows overlapping
            with `head` or `tail` are de-duplicated.
        :param random_state: random seed for the ``sample`` argument.
        :param lazy: if True, lazily evaluates dataframe against all validation
            checks and raises a ``SchemaErrors``. Otherwise, raise
            ``SchemaError`` as soon as one occurs.
        :param inplace: if True, applies coercion to the object of validation,
            otherwise creates a copy of the data.
        :returns: validated DataFrame or Series.
        """
        return self.get_backend(check_obj).validate(
            check_obj,
            schema=self,
            head=head,
            tail=tail,
            sample=sample,
            random_state=random_state,
            lazy=lazy,
            inplace=inplace,
        )

    def __repr__(self):
        return (
            f"<Schema {self.__class__.__name__}("
            f"indexes={self.indexes}, "
            f"coerce={self.coerce}, "
            f"strict={self.strict}, "
            f"name={self.name}, "
            f"ordered={self.ordered}"
            ")>"
        )

    def __str__(self):
        indent = " " * 4

        indexes_str = "[\n"
        for index in self.indexes:
            indexes_str += f"{indent * 2}{index}\n"
        indexes_str += f"{indent}]"

        return (
            f"<Schema {self.__class__.__name__}(\n"
            f"{indent}indexes={indexes_str}\n"
            f"{indent}coerce={self.coerce},\n"
            f"{indent}strict={self.strict},\n"
            f"{indent}name={self.name},\n"
            f"{indent}ordered={self.ordered}\n"
            ")>"
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    ###########################
    # Schema Strategy Methods #
    ###########################

    @st.strategy_import_error
    # NOTE: remove these ignore statements as part of
    # https://github.com/pandera-dev/pandera/issues/403
    # pylint: disable=arguments-differ
    def strategy(self, *, size=None):  # type: ignore
        return st.multiindex_strategy(indexes=self.indexes, size=size)

    # NOTE: remove these ignore statements as part of
    # https://github.com/pandera-dev/pandera/issues/403
    # pylint: disable=arguments-differ
    def example(self, size=None) -> pd.MultiIndex:  # type: ignore
        # pylint: disable=import-outside-toplevel,cyclic-import,import-error
        import hypothesis

        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore",
                category=hypothesis.errors.NonInteractiveExampleWarning,
            )
            return self.strategy(size=size).example()


def is_valid_multiindex_key(x: Tuple[Any, ...]) -> bool:
    """Check that a multi-index tuple key has all string elements"""
    return isinstance(x, tuple) and all(isinstance(i, str) for i in x)



"""Pandas backend utilities."""

from typing import Union

from pandera.dtypes import UniqueSettings


def convert_uniquesettings(unique: UniqueSettings) -> Union[bool, str]:
    """
    Converts UniqueSettings object to string that can be passed onto pandas .duplicated() call
    """
    # Default `keep` argument for pandas .duplicated() function
    keep_argument: Union[bool, str]
    if unique == "exclude_first":
        keep_argument = "first"
    elif unique == "exclude_last":
        keep_argument = "last"
    elif unique == "all":
        keep_argument = False
    else:
        raise ValueError(
            str(unique) + " is not a recognized report_duplicates value"
        )
    return keep_argument



"""Module to extract schema statsitics from schema objects."""

from pandera.schema_statistics.pandas import (
    infer_dataframe_statistics,
    infer_series_statistics,
    infer_index_statistics,
    parse_check_statistics,
    get_dataframe_schema_statistics,
    get_index_schema_statistics,
    get_series_schema_statistics,
    parse_checks,
)



"""Pandas backend implementation for schemas and checks."""

import pandas as pd

import pandera.typing
from pandera.api.checks import Check
from pandera.api.hypotheses import Hypothesis
from pandera.api.pandas.array import SeriesSchema
from pandera.api.pandas.container import DataFrameSchema
from pandera.api.pandas.components import Column, Index, MultiIndex

from pandera.backends.pandas import builtin_checks, builtin_hypotheses
from pandera.backends.pandas.checks import PandasCheckBackend
from pandera.backends.pandas.hypotheses import PandasHypothesisBackend
from pandera.backends.pandas.array import SeriesSchemaBackend
from pandera.backends.pandas.container import DataFrameSchemaBackend
from pandera.backends.pandas.components import (
    ColumnBackend,
    IndexBackend,
    MultiIndexBackend,
)


dataframe_datatypes = [pd.DataFrame]
series_datatypes = [pd.Series]
index_datatypes = [pd.Index]
multiindex_datatypes = [pd.MultiIndex]

if pandera.typing.dask.DASK_INSTALLED:
    import dask.dataframe as dd

    dataframe_datatypes.append(dd.DataFrame)
    series_datatypes.append(dd.Series)
    index_datatypes.append(dd.Index)

if pandera.typing.modin.MODIN_INSTALLED:
    import modin.pandas as mpd

    dataframe_datatypes.append(mpd.DataFrame)
    series_datatypes.append(mpd.Series)
    index_datatypes.append(mpd.Index)
    multiindex_datatypes.append(mpd.MultiIndex)

if pandera.typing.pyspark.PYSPARK_INSTALLED:
    import pyspark.pandas as ps

    dataframe_datatypes.append(ps.DataFrame)
    series_datatypes.append(ps.Series)
    index_datatypes.append(ps.Index)
    multiindex_datatypes.append(ps.MultiIndex)

if pandera.typing.geopandas.GEOPANDAS_INSTALLED:
    import geopandas as gpd

    dataframe_datatypes.append(gpd.GeoDataFrame)
    series_datatypes.append(gpd.GeoSeries)

for t in [
    *dataframe_datatypes,
    *series_datatypes,
    *index_datatypes,
]:
    Check.register_backend(t, PandasCheckBackend)
    Hypothesis.register_backend(t, PandasHypothesisBackend)

for t in dataframe_datatypes:
    DataFrameSchema.register_backend(t, DataFrameSchemaBackend)
    # SeriesSchema.register_backend(t, SeriesSchemaBackend)
    Column.register_backend(t, ColumnBackend)
    MultiIndex.register_backend(t, MultiIndexBackend)
    Index.register_backend(t, IndexBackend)

for t in series_datatypes:
    # DataFrameSchema.register_backend(t, DataFrameSchemaBackend)
    SeriesSchema.register_backend(t, SeriesSchemaBackend)
    Column.register_backend(t, ColumnBackend)
    Index.register_backend(t, IndexBackend)

for t in index_datatypes:
    Index.register_backend(t, IndexBackend)

for t in multiindex_datatypes:
    MultiIndex.register_backend(t, MultiIndexBackend)



"""Class-based dataframe model API configuration."""

from typing import Any, Optional


class BaseModelConfig:  # pylint:disable=R0903
    """Model configuration base class."""

    #: datatype of the data container. This overrides the data types specified
    #: in any of the fields.
    dtype: Optional[Any] = None

    name: Optional[str] = None  #: name of schema
    title: Optional[str] = None  #: human-readable label for schema
    description: Optional[str] = None  #: arbitrary textual description
    coerce: bool = False  #: coerce types of all schema components



"""Data types engine interface."""
# https://github.com/PyCQA/pylint/issues/3268
# pylint:disable=no-value-for-parameter
import functools
import inspect
import sys
from abc import ABCMeta
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    get_type_hints,
)

import typing_inspect

from pandera.dtypes import DataType


# register different TypedDict type depending on python version
if sys.version_info >= (3, 9):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict  # noqa


_DataType = TypeVar("_DataType", bound=DataType)
_Engine = TypeVar("_Engine", bound="Engine")
_EngineType = Type[_Engine]


if TYPE_CHECKING:  # pragma: no cover

    class Dispatch:
        """Only used for type annotation."""

        def __call__(self, data_type: Any, **kwds: Any) -> Any:
            pass

        @staticmethod
        def register(
            data_type: Any, func: Callable[[Any], DataType]
        ) -> Callable[[Any], DataType]:
            """Register a new implementation for the given cls."""

else:
    Dispatch = Callable[[Any], DataType]


def _is_typeddict(x: Type) -> bool:
    return x.__class__.__name__ == "_TypedDictMeta"


def _is_namedtuple(x: Type) -> bool:
    return tuple in getattr(x, "__bases__", ()) and hasattr(
        x, "__annotations__"
    )


@dataclass
class _DtypeRegistry:
    dispatch: Dispatch
    equivalents: Dict[Any, DataType]


class Engine(ABCMeta):
    """Base Engine metaclass.

    Keep a registry of concrete Engines.
    """

    _registry: Dict["Engine", _DtypeRegistry] = {}
    _registered_dtypes: Set[Type[DataType]]
    _base_pandera_dtypes: Tuple[Type[DataType]]

    def __new__(mcs, name, bases, namespace, **kwargs):
        base_pandera_dtypes = kwargs.pop("base_pandera_dtypes")
        try:
            namespace["_base_pandera_dtypes"] = tuple(base_pandera_dtypes)
        except TypeError:
            namespace["_base_pandera_dtypes"] = (base_pandera_dtypes,)

        namespace["_registered_dtypes"] = set()
        engine = super().__new__(mcs, name, bases, namespace, **kwargs)

        @functools.singledispatch
        def dtype(data_type: Any) -> DataType:
            raise ValueError(f"Data type '{data_type}' not understood")

        mcs._registry[engine] = _DtypeRegistry(dispatch=dtype, equivalents={})
        return engine

    def _check_source_dtype(cls, data_type: Any) -> None:
        if isinstance(data_type, cls._base_pandera_dtypes) or (
            inspect.isclass(data_type)
            and issubclass(data_type, cls._base_pandera_dtypes)
        ):
            base_names = [
                f"{base.__module__}.{base.__qualname__}"
                for base in cls._base_pandera_dtypes
            ]
            raise ValueError(
                f"Subclasses of {base_names} cannot be registered"
                f" with {cls.__name__}."
            )

    def _register_from_parametrized_dtype(
        cls,
        pandera_dtype_cls: Type[DataType],
    ) -> None:
        method = pandera_dtype_cls.__dict__["from_parametrized_dtype"]
        if not isinstance(method, classmethod):
            raise ValueError(
                f"{pandera_dtype_cls.__name__}.from_parametrized_dtype "
                + "must be a classmethod."
            )
        func = method.__func__
        annotations = get_type_hints(func).values()
        dtype = next(iter(annotations))  # get 1st annotation
        # parse typing.Union
        dtypes = typing_inspect.get_args(dtype) or [dtype]

        def _method(*args, **kwargs):
            return func(pandera_dtype_cls, *args, **kwargs)

        for source_dtype in dtypes:
            cls._check_source_dtype(source_dtype)
            cls._registry[cls].dispatch.register(source_dtype, _method)

    def _register_equivalents(
        cls, pandera_dtype_cls: Type[DataType], *source_dtypes: Any
    ) -> None:
        pandera_dtype = pandera_dtype_cls()  # type: ignore
        for source_dtype in source_dtypes:
            cls._check_source_dtype(source_dtype)
            cls._registry[cls].equivalents[source_dtype] = pandera_dtype

    def register_dtype(
        cls: _EngineType,
        pandera_dtype_cls: Type[_DataType] = None,
        *,
        equivalents: Optional[List[Any]] = None,
    ) -> Callable:
        """Register a Pandera :class:`~pandera.dtypes.DataType` with the engine,
        as class decorator.

        :param pandera_dtype: The DataType to register.
        :param equivalents: Equivalent scalar data type classes or
            non-parametrized data type instances.

        .. note::
            The classmethod ``from_parametrized_dtype`` will also be
            registered. See :ref:`here<dtypes>` for more usage details.

        :example:

        >>> import pandera as pa
        >>>
        >>> class MyDataType(pa.DataType):
        ...     pass
        >>>
        >>> class MyEngine(
        ...     metaclass=pa.engines.engine.Engine,
        ...     base_pandera_dtypes=MyDataType,
        ... ):
        ...     pass
        >>>
        >>> @MyEngine.register_dtype(equivalents=[bool])
        ... class MyBool(MyDataType):
        ...     pass

        """

        def _wrapper(pandera_dtype_cls: Type[_DataType]) -> Type[_DataType]:
            if not inspect.isclass(pandera_dtype_cls):
                raise ValueError(
                    f"{cls.__name__}.register_dtype can only decorate a class,"
                    f" got {pandera_dtype_cls}"
                )

            if equivalents:
                cls._register_equivalents(pandera_dtype_cls, *equivalents)

            if "from_parametrized_dtype" in pandera_dtype_cls.__dict__:
                cls._register_from_parametrized_dtype(pandera_dtype_cls)

            cls._registered_dtypes.add(pandera_dtype_cls)
            return pandera_dtype_cls

        if pandera_dtype_cls:
            return _wrapper(pandera_dtype_cls)

        return _wrapper

    def dtype(cls: _EngineType, data_type: Any) -> DataType:
        """Convert input into a Pandera :class:`DataType` object."""
        # pylint: disable=too-many-return-statements
        if isinstance(data_type, cls._base_pandera_dtypes):
            return data_type

        if inspect.isclass(data_type) and issubclass(
            data_type, cls._base_pandera_dtypes
        ):
            try:
                return data_type()
            except (TypeError, AttributeError) as err:
                raise TypeError(
                    f"DataType '{data_type.__name__}' cannot be instantiated: "
                    f"{err}\n "
                    + "Usage Tip: Use an instance or a string representation."
                ) from err

        registry = cls._registry[cls]

        # handle python generic types, e.g. typing.Dict[str, str]
        datatype_origin = typing_inspect.get_origin(data_type)
        if datatype_origin is not None:
            equivalent_data_type = registry.equivalents.get(datatype_origin)
            return type(equivalent_data_type)(data_type)

        # handle python's special declared type constructs like NamedTuple and
        # TypedDict
        datatype_generic_bases = (
            # handle python < 3.9 cases, where TypedDict/NameDtuple isn't part
            # of the generic base classes returned by
            # typing_inspect.get_generic_bases
            ((TypedDict,) if _is_typeddict(data_type) else ())
            or ((NamedTuple,) if _is_namedtuple(data_type) else ())
            or typing_inspect.get_generic_bases(data_type)
        )
        if datatype_generic_bases:
            equivalent_data_type = None
            for base in datatype_generic_bases:
                equivalent_data_type = registry.equivalents.get(base)
                break
            if equivalent_data_type is None:
                raise TypeError(
                    f"Type '{data_type}' not understood by {cls.__name__}."
                )
            return type(equivalent_data_type)(data_type)

        equivalent_data_type = registry.equivalents.get(data_type)
        if equivalent_data_type is not None:
            return equivalent_data_type
        elif isinstance(data_type, DataType):
            # in the case where data_type is a parameterized dtypes.DataType instance that isn't
            # in the equivalents registry, use its type to get the equivalent, and feed
            # the parameters into the recognized data type class.
            equivalent_data_type = registry.equivalents.get(type(data_type))
            if equivalent_data_type is not None:
                return type(equivalent_data_type)(**data_type.__dict__)

        try:
            return registry.dispatch(data_type)
        except (KeyError, ValueError):
            raise TypeError(
                f"Data type '{data_type}' not understood by {cls.__name__}."
            ) from None

    def get_registered_dtypes(  # pylint:disable=W1401
        cls,
    ) -> List[Type[DataType]]:
        r"""Return the :class:`pandera.dtypes.DataType`\s registered
        with this engine."""
        return list(cls._registered_dtypes)



"""Configuration for external packages."""

import os

try:
    # try importing pyspark to see if it exists. This is important because the
    # pandera.typing module defines a Series type that inherits from
    # pandas.Series, and pyspark v1+ injects a __getitem__ method to pandas
    # Series and DataFrames to support type hinting:
    # https://spark.apache.org/docs/3.2.0/api/python/user_guide/pandas_on_spark/typehints.html#type-hinting-with-names
    # pylint: disable=unused-import
    if os.getenv("SPARK_LOCAL_IP") is None:
        os.environ["SPARK_LOCAL_IP"] = "127.0.0.1"
    if os.getenv("PYARROW_IGNORE_TIMEZONE") is None:
        # This can be overriden by the user
        os.environ["PYARROW_IGNORE_TIMEZONE"] = "1"

    import pyspark.pandas
except ImportError:
    os.environ.pop("SPARK_LOCAL_IP")
    os.environ.pop("PYARROW_IGNORE_TIMEZONE")



"""Pandera type annotations for Dask."""

from typing import TYPE_CHECKING, Generic, TypeVar

from packaging import version

from pandera.typing.common import DataFrameBase, IndexBase, SeriesBase
from pandera.typing.pandas import GenericDtype, DataFrameModel

try:
    import modin
    import modin.pandas as mpd

    MODIN_INSTALLED = True
except ImportError:
    MODIN_INSTALLED = False


def modin_version():
    """Return the modin version."""
    return version.parse(modin.__version__)


# pylint:disable=invalid-name
if TYPE_CHECKING:
    T = TypeVar("T")  # pragma: no cover
else:
    T = DataFrameModel


if MODIN_INSTALLED:

    # pylint: disable=too-few-public-methods
    class DataFrame(DataFrameBase, mpd.DataFrame, Generic[T]):
        """
        Representation of dask.dataframe.DataFrame, only used for type
        annotation.

        *new in 0.8.0*
        """

    # pylint:disable=too-few-public-methods,abstract-method
    class Series(SeriesBase, mpd.Series, Generic[GenericDtype]):
        """Representation of pandas.Series, only used for type annotation.

        *new in 0.8.0*
        """

    # pylint:disable=too-few-public-methods,abstract-method
    class Index(IndexBase, mpd.Index, Generic[GenericDtype]):
        """Representation of pandas.Index, only used for type annotation.

        *new in 0.8.0*
        """



"""Base functions for Parsing, Validation, and Error Reporting Backends.

This class should implement a common interface of operations needed for
data validation. These operations are exposed as methods that are composed
together to implement the pandera schema specification.
"""

import inspect
from abc import ABC
from functools import wraps
from typing import Any, Dict, Tuple, Type, Union

from pandera.backends.base import BaseSchemaBackend
from pandera.errors import BackendNotFoundError
from pandera.dtypes import DataType

DtypeInputTypes = Union[str, type, DataType, Type]


class BaseSchema(ABC):
    """Core schema specification."""

    BACKEND_REGISTRY: Dict[
        Tuple[Type, Type], Type[BaseSchemaBackend]
    ] = {}  # noqa

    def __init__(
        self,
        dtype=None,
        checks=None,
        coerce=False,
        name=None,
        title=None,
        description=None,
    ):
        """Abstract base schema initializer."""
        self.dtype = dtype
        self.checks = checks
        self.coerce = coerce
        self.name = name
        self.title = title
        self.description = description

    def validate(
        self,
        check_obj,
        head=None,
        tail=None,
        sample=None,
        random_state=None,
        lazy=False,
        inplace=False,
    ):
        """Validate method to be implemented by subclass."""
        raise NotImplementedError

    def coerce_dtype(self, check_obj):
        """Coerce object to the expected type."""
        raise NotImplementedError

    @property
    def properties(self):
        """Get the properties of the schema for serialization purposes."""
        raise NotImplementedError

    @classmethod
    def register_backend(cls, type_: Type, backend: Type[BaseSchemaBackend]):
        """Register a schema backend for this class."""
        cls.BACKEND_REGISTRY[(cls, type_)] = backend

    @classmethod
    def get_backend(cls, check_obj: Any) -> BaseSchemaBackend:
        """Get the backend associated with the type of ``check_obj`` ."""
        check_obj_cls = type(check_obj)
        classes = inspect.getmro(check_obj_cls)
        for _class in classes:
            try:
                return cls.BACKEND_REGISTRY[(cls, _class)]()
            except KeyError:
                pass
        raise BackendNotFoundError(
            f"Backend not found for backend, class: {(cls, check_obj_cls)}. "
            f"Looked up the following base classes: {classes}"
        )


def inferred_schema_guard(method):
    """
    Invoking a method wrapped with this decorator will set _is_inferred to
    False.
    """

    @wraps(method)
    def wrapper(schema, *args, **kwargs):
        new_schema = method(schema, *args, **kwargs)
        if new_schema is not None and id(new_schema) != id(schema):
            # if method returns a copy of the schema object,
            # the original schema instance and the copy should be set to
            # not inferred.
            new_schema._is_inferred = False
        return new_schema

    return wrapper



"""Typing module.

For backwards compatibility, pandas types are exposed to the top-level scope of
the typing module.
"""

from typing import Set, Type

from pandera.typing import dask, fastapi, geopandas, modin, pyspark
from pandera.typing.common import (
    BOOL,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    AnnotationInfo,
    Bool,
    Category,
    Date,
    DateTime,
    Decimal,
    Float,
    Float16,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    Object,
    String,
    Timedelta,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)
from pandera.typing.pandas import DataFrame, Index, Series

DATAFRAME_TYPES: Set[Type] = {DataFrame}
SERIES_TYPES: Set[Type] = {Series}
INDEX_TYPES: Set[Type] = {Index}

if dask.DASK_INSTALLED:
    DATAFRAME_TYPES.update({dask.DataFrame})
    SERIES_TYPES.update({dask.Series})
    INDEX_TYPES.update({dask.Index})

if modin.MODIN_INSTALLED:
    DATAFRAME_TYPES.update({modin.DataFrame})
    SERIES_TYPES.update({modin.Series})
    INDEX_TYPES.update({modin.Index})

if pyspark.PYSPARK_INSTALLED:
    DATAFRAME_TYPES.update({pyspark.DataFrame})
    SERIES_TYPES.update({pyspark.Series})
    INDEX_TYPES.update({pyspark.Index})  # type: ignore [arg-type]


if geopandas.GEOPANDAS_INSTALLED:
    DATAFRAME_TYPES.update({geopandas.GeoDataFrame})
    SERIES_TYPES.update({geopandas.GeoSeries})


__all__ = [
    "DataFrame",
    "Series",
    "Index",
]



# pylint: skip-file
# NOTE: skip file since py=3.10 yields these errors:
# https://github.com/pandera-dev/pandera/runs/4998710717?check_suite_focus=true
"""Register pyspark accessor for pandera schema metadata."""

from pyspark.pandas.extensions import (
    register_dataframe_accessor,
    register_series_accessor,
)

from pandera.accessors.pandas_accessor import (
    PanderaDataFrameAccessor,
    PanderaSeriesAccessor,
)

register_dataframe_accessor("pandera")(PanderaDataFrameAccessor)
register_series_accessor("pandera")(PanderaSeriesAccessor)



"""Module for inferring the statistics of pandas objects."""
import warnings
from typing import Any, Dict, Union

import pandas as pd

from pandera import dtypes
from pandera.api.checks import Check
from pandera.engines import pandas_engine


def infer_dataframe_statistics(df: pd.DataFrame) -> Dict[str, Any]:
    """Infer column and index statistics from a pandas DataFrame."""
    nullable_columns = df.isna().any()
    inferred_column_dtypes = {col: _get_array_type(df[col]) for col in df}
    column_statistics = {
        col: {
            "dtype": dtype,
            "nullable": bool(nullable_columns[col]),  # type: ignore
            "checks": _get_array_check_statistics(df[col], dtype),
        }
        for col, dtype in inferred_column_dtypes.items()
    }
    return {
        "columns": column_statistics if column_statistics else None,
        "index": infer_index_statistics(df.index),
    }


def infer_series_statistics(series: pd.Series) -> Dict[str, Any]:
    """Infer column and index statistics from a pandas Series."""
    dtype = _get_array_type(series)
    return {
        "dtype": dtype,
        "nullable": bool(series.isna().any()),
        "checks": _get_array_check_statistics(series, dtype),
        "name": series.name,
    }


def infer_index_statistics(index: Union[pd.Index, pd.MultiIndex]):
    """Infer index statistics given a pandas Index object."""

    def _index_stats(index_level):
        dtype = _get_array_type(index_level)
        return {
            "dtype": dtype,
            "nullable": bool(index_level.isna().any()),
            "checks": _get_array_check_statistics(index_level, dtype),
            "name": index_level.name,
        }

    if isinstance(index, pd.MultiIndex):
        index_statistics = [
            _index_stats(index.get_level_values(i))
            for i in range(index.nlevels)
        ]
    elif isinstance(index, pd.Index):
        index_statistics = [_index_stats(index)]
    else:
        warnings.warn(
            f"index type {type(index)} not recognized, skipping index inference",
            UserWarning,
        )
        index_statistics = []
    return index_statistics if index_statistics else None


def parse_check_statistics(check_stats: Union[Dict[str, Any], None]):
    """Convert check statistics to a list of Check objects."""
    if check_stats is None:
        return None
    checks = []
    for check_name, stats in check_stats.items():
        check = getattr(Check, check_name)
        try:
            checks.append(check(**stats))
        except TypeError:
            # if stats cannot be unpacked as key-word args, assume unary check.
            checks.append(check(stats))
    return checks if checks else None


def get_dataframe_schema_statistics(dataframe_schema):
    """Get statistical properties from dataframe schema."""
    statistics = {
        "columns": {
            col_name: {
                "dtype": column.dtype,
                "nullable": column.nullable,
                "coerce": column.coerce,
                "required": column.required,
                "regex": column.regex,
                "checks": parse_checks(column.checks),
                "unique": column.unique,
                "description": column.description,
                "title": column.title,
            }
            for col_name, column in dataframe_schema.columns.items()
        },
        "checks": parse_checks(dataframe_schema.checks),
        "index": (
            None
            if dataframe_schema.index is None
            else get_index_schema_statistics(dataframe_schema.index)
        ),
        "coerce": dataframe_schema.coerce,
    }
    return statistics


def _get_series_base_schema_statistics(series_schema_base):
    return {
        "dtype": series_schema_base.dtype,
        "nullable": series_schema_base.nullable,
        "checks": parse_checks(series_schema_base.checks),
        "coerce": series_schema_base.coerce,
        "name": series_schema_base.name,
        "unique": series_schema_base.unique,
        "title": series_schema_base.title,
        "description": series_schema_base.description,
    }


def get_index_schema_statistics(index_schema_component):
    """Get statistical properties of index schema component."""
    try:
        # get index components from MultiIndex
        index_components = index_schema_component.indexes
    except AttributeError:
        index_components = [index_schema_component]
    return [
        _get_series_base_schema_statistics(index_component)
        for index_component in index_components
    ]


def get_series_schema_statistics(series_schema):
    """Get statistical properties from series schema."""
    return _get_series_base_schema_statistics(series_schema)


def parse_checks(checks) -> Union[Dict[str, Any], None]:
    """Convert Check object to check statistics."""
    check_statistics = {}
    _check_memo = {}
    for check in checks:
        if check not in Check:
            warnings.warn(
                "Only registered checks may be serialized to statistics. "
                "Did you forget to register it with the extension API? "
                f"Check `{check.name}` will be skipped."
            )
            continue

        check_statistics[check.name] = (
            {} if check.statistics is None else check.statistics
        )
        _check_memo[check.name] = check

    # raise ValueError on incompatible checks
    if (
        "greater_than_or_equal_to" in check_statistics
        and "less_than_or_equal_to" in check_statistics
    ):
        min_value = check_statistics.get(
            "greater_than_or_equal_to", float("-inf")
        )["min_value"]
        max_value = check_statistics.get(
            "less_than_or_equal_to", float("inf")
        )["max_value"]
        if min_value > max_value:
            raise ValueError(
                f"checks {_check_memo['greater_than_or_equal_to']} "
                f"and {_check_memo['less_than_or_equal_to']} are incompatible, reason: "
                f"min value {min_value} > max value {max_value}"
            )
    return check_statistics if check_statistics else None


def _get_array_type(x):
    # get most granular type possible

    data_type = pandas_engine.Engine.dtype(x.dtype)
    # for object arrays, try to infer dtype
    if data_type is pandas_engine.Engine.dtype("object"):
        inferred_alias = pd.api.types.infer_dtype(x, skipna=False)
        if inferred_alias != "string":
            data_type = pandas_engine.Engine.dtype(inferred_alias)
    return data_type


def _get_array_check_statistics(
    x, data_type: dtypes.DataType
) -> Union[Dict[str, Any], None]:
    """Get check statistics from an array-like object."""
    if x.isna().all():
        return None
    if dtypes.is_datetime(data_type):
        check_stats = {
            "greater_than_or_equal_to": x.min(),
            "less_than_or_equal_to": x.max(),
        }
    elif dtypes.is_numeric(data_type) and not dtypes.is_bool(data_type):
        check_stats = {
            "greater_than_or_equal_to": float(x.min()),
            "less_than_or_equal_to": float(x.max()),
        }
    elif dtypes.is_category(data_type):
        try:
            categories = x.cat.categories
        except AttributeError:
            categories = x.categories
        check_stats = {
            "isin": categories.tolist(),
        }
    else:
        check_stats = {}
    return check_stats if check_stats else None



"""Register pandas accessor for pandera schema metadata."""

from typing import Optional, Union

import pandas as pd

from pandera.api.pandas.array import SeriesSchema
from pandera.api.pandas.container import DataFrameSchema

Schemas = Union[DataFrameSchema, SeriesSchema]


class PanderaAccessor:
    """Pandera accessor for pandas object."""

    def __init__(self, pandas_obj):
        """Initialize the pandera accessor."""
        self._pandas_obj = pandas_obj
        self._schema: Optional[Schemas] = None

    @staticmethod
    def check_schema_type(schema: Schemas):
        """Abstract method for checking the schema type."""
        raise NotImplementedError

    def add_schema(self, schema):
        """Add a schema to the pandas object."""
        self.check_schema_type(schema)
        self._schema = schema
        return self._pandas_obj

    @property
    def schema(self) -> Optional[Schemas]:
        """Access schema metadata."""
        return self._schema


@pd.api.extensions.register_dataframe_accessor("pandera")
class PanderaDataFrameAccessor(PanderaAccessor):
    """Pandera accessor for pandas DataFrame."""

    @staticmethod
    def check_schema_type(schema):
        if not isinstance(schema, DataFrameSchema):
            raise TypeError(
                f"schema arg must be a DataFrameSchema, found {type(schema)}"
            )


@pd.api.extensions.register_series_accessor("pandera")
class PanderaSeriesAccessor(PanderaAccessor):
    """Pandera accessor for pandas Series."""

    @staticmethod
    def check_schema_type(schema):
        if not isinstance(schema, SeriesSchema):
            raise TypeError(
                f"schema arg must be a SeriesSchema, found {type(schema)}"
            )



"""Decorators for integrating pandera into existing data pipelines."""
from inspect import ismethod
from typing import Callable


def _is_like_classmethod(fn: Callable) -> bool:
    """A regular method defined on a metaclass behaves the same way as
    a method decorated with @classmethod defined on a regular class.

    This function covers both use cases.
    """
    is_method = ismethod(fn)
    return is_method and isinstance(fn.__self__, type)  # type: ignore[attr-defined]


def is_decorated_classmethod(fn: Callable) -> bool:
    """Check if fn is a classmethod declared with the @classmethod decorator.

    Adapted from:
    https://stackoverflow.com/questions/19227724/check-if-a-function-uses-classmethod
    """
    if not _is_like_classmethod(fn):
        return False
    bound_to = fn.__self__  # type: ignore[attr-defined]
    assert isinstance(bound_to, type)
    name = fn.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False


def is_classmethod_from_meta(fn: Callable) -> bool:
    """Check if fn is a regular method defined on a metaclass
    (which behaves like an @classmethod method defined on a regular class)."""
    return not is_decorated_classmethod(fn) and _is_like_classmethod(fn)



"""Utility functions for pandas validation."""

from functools import lru_cache
from typing import List, NamedTuple, Tuple, Type, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore [misc]

import numpy as np
import pandas as pd

from pandera.api.checks import Check
from pandera.api.hypotheses import Hypothesis
from pandera.dtypes import DataType

CheckList = Union[Check, List[Union[Check, Hypothesis]]]

PandasDtypeInputTypes = Union[
    str,
    type,
    DataType,
    Type,
    pd.core.dtypes.base.ExtensionDtype,
    np.dtype,
]

StrictType = Union[bool, Literal["filter"]]

SupportedTypes = NamedTuple(
    "SupportedTypes",
    (
        ("table_types", Tuple[type, ...]),
        ("field_types", Tuple[type, ...]),
        ("index_types", Tuple[type, ...]),
        ("multiindex_types", Tuple[type, ...]),
    ),
)


@lru_cache(maxsize=None)
def supported_types() -> SupportedTypes:
    """Get the types supported by pandera schemas."""
    # pylint: disable=import-outside-toplevel
    table_types = [pd.DataFrame]
    field_types = [pd.Series]
    index_types = [pd.Index]
    multiindex_types = [pd.MultiIndex]

    try:
        import pyspark.pandas as ps

        table_types.append(ps.DataFrame)
        field_types.append(ps.Series)
        index_types.append(ps.Index)
        multiindex_types.append(ps.MultiIndex)
    except ImportError:
        pass
    try:  # pragma: no cover
        import modin.pandas as mpd

        table_types.append(mpd.DataFrame)
        field_types.append(mpd.Series)
        index_types.append(mpd.Index)
        multiindex_types.append(mpd.MultiIndex)
    except ImportError:
        pass
    try:
        import dask.dataframe as dd

        table_types.append(dd.DataFrame)
        field_types.append(dd.Series)
        index_types.append(dd.Index)
    except ImportError:
        pass

    return SupportedTypes(
        tuple(table_types),
        tuple(field_types),
        tuple(index_types),
        tuple(multiindex_types),
    )


def is_table(obj):
    """Verifies whether an object is table-like.

    Where a table is a 2-dimensional data matrix of rows and columns, which
    can be indexed in multiple different ways.
    """
    return isinstance(obj, supported_types().table_types)


def is_field(obj):
    """Verifies whether an object is field-like.

    Where a field is a columnar representation of data in a table-like
    data structure.
    """
    return isinstance(obj, supported_types().field_types)


def is_index(obj):
    """Verifies whether an object is a table index."""
    return isinstance(obj, supported_types().index_types)


def is_multiindex(obj):
    """Verifies whether an object is a multi-level table index."""
    return isinstance(obj, supported_types().multiindex_types)


def is_table_or_field(obj):
    """Verifies whether an object is table- or field-like."""
    return is_table(obj) or is_field(obj)


is_supported_check_obj = is_table_or_field


def is_bool(x):
    """Verifies whether an object is a boolean type."""
    return isinstance(x, (bool, np.bool_))



"""Pandera type annotations for GeoPandas."""
from typing import TYPE_CHECKING, Generic, TypeVar

from pandera.typing.common import DataFrameBase, SeriesBase

from pandera.typing.pandas import DataFrameModel

try:
    import geopandas as gpd

    GEOPANDAS_INSTALLED = True
except ImportError:  # pragma: no cover
    GEOPANDAS_INSTALLED = False


if GEOPANDAS_INSTALLED:
    # pylint: disable=import-outside-toplevel,ungrouped-imports
    from pandera.engines.pandas_engine import Geometry

    # pylint:disable=invalid-name
    if TYPE_CHECKING:
        T = TypeVar("T")  # pragma: no cover
    else:
        T = DataFrameModel

    # pylint:disable=too-few-public-methods
    class GeoSeries(SeriesBase, gpd.GeoSeries, Generic[T]):
        """
        Representation of geopandas.GeoSeries, only used for type annotation.
        """

        default_dtype = Geometry

    class GeoDataFrame(DataFrameBase, gpd.GeoDataFrame, Generic[T]):
        """
        A generic type for geopandas.GeoDataFrame.
        """



"""Data validation base check."""

import inspect
from itertools import chain
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
)
import pandas as pd
from multimethod import multidispatch as _multidispatch

from pandera.backends.base import BaseCheckBackend


class CheckResult(NamedTuple):
    """Check result for user-defined checks."""

    check_output: Any
    check_passed: bool
    checked_object: Any
    failure_cases: Any


GroupbyObject = Union[
    pd.core.groupby.generic.SeriesGroupBy,
    pd.core.groupby.generic.DataFrameGroupBy,
]

SeriesCheckObj = Union[pd.Series, Dict[str, pd.Series]]
DataFrameCheckObj = Union[pd.DataFrame, Dict[str, pd.DataFrame]]


_T = TypeVar("_T", bound="BaseCheck")


# pylint: disable=invalid-name
class multidispatch(_multidispatch):
    """
    Custom multidispatch class to handle copy, deepcopy, and code retrieval.
    """

    @property
    def __code__(self):
        """Retrieves the 'base' function of the multidispatch object."""
        assert (
            len(self) > 0
        ), f"multidispatch object {self} has no functions registered"
        fn, *_ = [*self.values()]  # type: ignore[misc]
        return fn.__code__

    def __reduce__(self):
        """
        Handle custom pickling reduction method by initializing a new
        multidispatch object, wrapped with the base function.
        """
        state = self.__dict__
        # make sure all registered functions at time of pickling are captured
        state["__registered_functions__"] = [*self.values()]
        return (
            multidispatch,  # object creation function
            (state["__wrapped__"],),  # arguments to said function
            state,  # arguments to `__setstate__` after creation
        )

    def __setstate__(self, state):
        """Custom unpickling logic."""
        self.__dict__ = state
        # rehydrate the multidispatch object with unpickled registered functions
        for fn in state["__registered_functions__"]:
            self.register(fn)


class MetaCheck(type):  # pragma: no cover
    """Check metaclass."""

    BACKEND_REGISTRY: Dict[
        Tuple[Type, Type], Type[BaseCheckBackend]
    ] = {}  # noqa
    """Registry of check backends implemented for specific data objects."""

    CHECK_FUNCTION_REGISTRY: Dict[str, Callable] = {}  # noqa
    """Built-in check function registry."""

    REGISTERED_CUSTOM_CHECKS: Dict[str, Callable] = {}  # noqa
    """User-defined custom checks."""

    def __getattr__(cls, name: str) -> Any:
        """Prevent attribute errors for registered checks."""
        attr = {
            **cls.__dict__,
            **cls.CHECK_FUNCTION_REGISTRY,
            **cls.REGISTERED_CUSTOM_CHECKS,
        }.get(name)
        if attr is None:
            raise AttributeError(
                f"'{cls}' object has no attribute '{name}'. "
                "Make sure any custom checks have been registered "
                "using the extensions api."
            )
        return attr

    def __dir__(cls) -> Iterable[str]:
        """Allow custom checks to show up as attributes when autocompleting."""
        return chain(
            super().__dir__(),
            cls.CHECK_FUNCTION_REGISTRY.keys(),
            cls.REGISTERED_CUSTOM_CHECKS.keys(),
        )

    # pylint: disable=line-too-long
    # mypy has limited metaclass support so this doesn't pass typecheck
    # see https://mypy.readthedocs.io/en/stable/metaclasses.html#gotchas-and-limitations-of-metaclass-support
    # pylint: enable=line-too-long
    @no_type_check
    def __contains__(cls: Type[_T], item: Union[_T, str]) -> bool:
        """Allow lookups for registered checks."""
        if isinstance(item, cls):
            name = item.name
            return hasattr(cls, name)

        # assume item is str
        return hasattr(cls, item)


class BaseCheck(metaclass=MetaCheck):
    """Check base class."""

    def __init__(
        self,
        name: Optional[str] = None,
        error: Optional[str] = None,
        statistics: Optional[Dict[str, Any]] = None,
    ):
        self.name = name
        self.error = error
        self.statistics = statistics

    @classmethod
    def register_builtin_check_fn(cls, fn: Callable):
        """Registers a built-in check function"""
        cls.CHECK_FUNCTION_REGISTRY[fn.__name__] = multidispatch(fn)
        return fn

    @classmethod
    def get_builtin_check_fn(cls, name: str):
        """Gets a built-in check function"""
        return cls.CHECK_FUNCTION_REGISTRY[name]

    @classmethod
    def from_builtin_check_name(
        cls,
        name: str,
        init_kwargs,
        error: Union[str, Callable],
        statistics: Dict[str, Any] = None,
        **check_kwargs,
    ):
        """Create a Check object from a built-in check's name."""
        kws = {**init_kwargs, **check_kwargs}
        if "error" not in kws:
            kws["error"] = error

        # statistics are the raw check constraint values that are untransformed
        # by the check object
        if statistics is None:
            statistics = check_kwargs

        return cls(
            cls.get_builtin_check_fn(name),
            statistics=statistics,
            **kws,
        )

    @classmethod
    def register_backend(cls, type_: Type, backend: Type[BaseCheckBackend]):
        """Register a backend for the specified type."""
        cls.BACKEND_REGISTRY[(cls, type_)] = backend

    @classmethod
    def get_backend(cls, check_obj: Any) -> Type[BaseCheckBackend]:
        """Get the backend associated with the type of ``check_obj`` ."""
        check_obj_cls = type(check_obj)
        classes = inspect.getmro(check_obj_cls)
        for _class in classes:
            try:
                return cls.BACKEND_REGISTRY[(cls, _class)]
            except KeyError:
                pass
        raise KeyError(
            f"Backend not found for class: {check_obj_cls}. Looked up the "
            f"following base classes: {classes}"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        are_check_fn_objects_equal = (
            self._get_check_fn_code() == other._get_check_fn_code()
        )

        try:
            are_strategy_fn_objects_equal = all(
                getattr(self.__dict__.get("strategy"), attr)
                == getattr(other.__dict__.get("strategy"), attr)
                for attr in ["func", "args", "keywords"]
            )
        except AttributeError:
            are_strategy_fn_objects_equal = True

        are_all_other_check_attributes_equal = {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["_check_fn", "strategy"]
        } == {
            k: v
            for k, v in other.__dict__.items()
            if k not in ["_check_fn", "strategy"]
        }

        return (
            are_check_fn_objects_equal
            and are_strategy_fn_objects_equal
            and are_all_other_check_attributes_equal
        )

    def _get_check_fn_code(self):
        check_fn = self.__dict__["_check_fn"]
        try:
            code = check_fn.__code__.co_code
        except AttributeError:
            # try accessing the functools.partial wrapper
            code = check_fn.func.__code__.co_code

        return code

    def __hash__(self) -> int:
        return hash(self._get_check_fn_code())

    def __repr__(self) -> str:
        return (
            f"<Check {self.name}: {self.error}>"
            if self.error is not None
            else f"<Check {self.name}>"
        )
